//---------------------------------------------------------------------------

#include "stdafx.h"

using namespace std;

#ifdef _MSC_VER
#pragma hdrstop
#endif

#include "myExcept.h"
#include "bitvector.h"
#include "PLA_type.h"
#include "Cube.h"
#include "SoPLA.h"
#include "utils.h"
//---------------------------------------------------------------------------


SoPLA::SoPLA(void)
{
	inps = 0;
	indexed = false;
	type = true;
	DFF = false;
}

SoPLA::SoPLA( const unsigned int iv )
{
	inps = iv;
	indexed = false;
	type = true;
	DFF = false;
}

SoPLA::SoPLA( const string s )
{
	inps = s.size();
	indexed = false;
	type = true;
	DFF = false;
	AppendTerm(s);
}

SoPLA::SoPLA( const SoPLA *_pla )
{
	inps = _pla->inps;
	inames = _pla->inames;
	name = _pla->name;
	type = _pla->type;
	DFF = _pla->DFF;
	indexed = false;
}

SoPLA::SoPLA( const unsigned int iv, const bool _type )
{
	inps = iv;
	indexed = false;
	type = _type;
	DFF = false;
}

void SoPLA::clearData(void)
{
	data.clear();
}

void SoPLA::clear(void)
// Clears all the PLA, deallocates all the structures
{
	data.clear();
    inames.clear();
	indexed = false;
}

SoPLA::~SoPLA(void)
{
}

void SoPLA::setInputs( const int i )
{
	inps = i;
}

list<Cube>::iterator SoPLA::AppendTerm( const Cube &t )
// Creates and appends the term. Returns the iterator pointing to this term
{
	list<Cube>::iterator ti;

	if ( inps == 0 ) inps = t.inputs();
	else if ( inps != t.inputs() ) throw RuntimeException( "SoPLA::AppendTerm: incompatible terms" );
	data.push_back(t);
	ti = data.end();
	ti--;
	return ti;
}

list<Cube>::iterator SoPLA::AppendTerm( const string s )
// Creates and appends the term. Returns the iterator pointing to this term
{
	Cube t;

	t.setCube(s);
	return AppendTerm(t);
}

void SoPLA::AppendPLA( const SoPLA &_pla )
// appends a given PLA to this. The argument PLA is copied
{
	CubeList::const_iterator ti;

	for ( ti = _pla.data.begin(); ti != _pla.data.end(); ++ti )
		AppendTerm(*ti);
}

void SoPLA::Splice( SoPLA &_pla )
// appends a given PLA to this and clears the argument pla
{
	data.splice(data.end(), _pla.data);
}

void SoPLA::Splice( CubeList &terms )
// appends a term list to this and clears the argument pla
{
	data.splice(data.end(), terms);
}

string SoPLA::getIName(const unsigned int i) const
{
	if ( inames.empty() ) return "";
	if ( i >= inps ) throw RuntimeException( "SoPLA::getIName: Incorrect index" );
	else return inames[i];
}

void SoPLA::setIName(const unsigned int i, const string name)
{
	if ( inames.empty() ) inames.resize( inps, "" );
	if ( i >= inps ) throw RuntimeException( "SoPLA::setIName: Incorrect index" );
	inames[i] = name;
}

bool SoPLA::empty(void) const
{
	return ( data.size() == 0 );
}

void SoPLA::Reindex(void)
// Reindexes the PLA, if not indexed
{
    list<Cube>::iterator ti;
    unsigned int i = 0;

    if ( indexed ) return;
	indexterm.clear();
	for ( ti = data.begin(); ti != data.end(); ++ti ) {
        indexterm.push_back(ti);
		ti->setIndex(i++);
	}
    indexed = true;
}

list<Cube>::iterator SoPLA::getTerm( const unsigned int i )
// Returns an iterator to an indexed term. If not indexed, Reindex is called
{
	if ( !indexed ) Reindex();
	return indexterm[i];
}

list<Cube>::iterator SoPLA::DeleteTerm(const list<Cube>::iterator ti)
// Deletes the term pointed by iterator. Returns an iterator to the following term
{
	indexed = false;
    return data.erase(ti);
}

void SoPLA::RemoveAbsorbed(void)
// Removes absorbed terms
{
    CubeList::iterator ti, ti1;
	bool deleted;

	if ( data.size() < 2 ) return;
	ti = data.begin();
	do {
		ti1 = ti;
		++ti1;
		deleted = false;
		while ( ti1 != data.end() ) {
			if ( ti->contains( *ti1 ) ) ti1 = DeleteTerm(ti1);
			else if ( ti1->contains(*ti) ) {
				ti = DeleteTerm(ti);
				deleted = true;
				break;
			} else ++ti1;
		}
		if ( !deleted ) ++ti;
	} while ( ti != data.end() );
}

unsigned int SoPLA::getLiterals(void)
// Returns the number of PLA literals
{
    unsigned int l = 0;
    list<Cube>::iterator ti;

    for ( ti = data.begin(); ti != data.end(); ++ti )
        l += inps-ti->getDim();
    return l;
}

unsigned int SoPLA::getGEs(void)
// Returns GEs*2
{
	unsigned int ge = 0;
	list<Cube>::iterator ti;
	
	for ( ti = data.begin(); ti != data.end(); ++ti )
		if ( inps-ti->getDim() > 1 ) ge += inps-ti->getDim();
	ge += getTerms();
	if ( getTerms() == 1 ) ge--;		// no OR gate
	if ( getTerms() == 1 && inps == 1 && getData()->front().getDim() == 0 ) ge++;  // inverter or buffer
	return ge;
}

void SoPLA::RemoveAllDCCols(void)
{
	list<Cube>::iterator ti;
	unsigned int i;
	vector<bool> alldc;
	string new_IM;
	vector<string> new_names;
	unsigned int cnt = 0;

	if ( inps == 1 ) return;
	alldc.resize( inps, false );
	for ( i = 0; i < inps; i++ ) {
		for ( ti = data.begin(); ti != data.end(); ++ti )
			if ( !ti->isLit_dc(i) ) break;
		if ( ti == data.end() ) {
			alldc[i] = true;
			cnt++;
		}
	}
	if ( cnt == 0 ) return;
	for ( ti = data.begin(); ti != data.end(); ++ti ) {
		new_IM.clear();
		for ( i = 0; i < inps; i++ )
			if ( !alldc[i] ) new_IM.push_back( ti->getLit(i) );
		ti->clear();
		ti->setCube(new_IM);
	}
	new_names.clear();
	for ( i = 0; i < inps; i++ )
		if ( !alldc[i] ) new_names.push_back( inames[i] );
	inames = new_names;
	setInputs(inputs()-cnt);
}

SoPLA *SoPLA::Cofactor( const Cube &t ) const
// Computes a cofactor of pla wrt. t. The resulting PLA is of the same dimensions as pla
{
	SoPLA *res;
	CubeList::const_iterator ti;
	Cube newterm;
	unsigned int i;

	if ( inps != t.inputs() ) throw RuntimeException( "SoPLA::Cofactor: PLA and the term are incompatible" );
	res = new SoPLA(this);
	for ( ti = getConstDataRef().begin(); ti != getConstDataRef().end(); ++ti ) {
		newterm = (*ti);
		for ( i = 0; i < inps; i++ )
			if ( !t.isLit_dc(i) ) {
				if ( t.getLit(i) != ti->getLit(i) && !ti->isLit_dc(i) ) break;
				else newterm.setLit_dc(i);
			}
		if ( i == inps ) res->AppendTerm(newterm);
	}
	return res;
}

bool SoPLA::isConstant(void) const
// is the PLA constant 0 or 1?
{
	unsigned int i;

	if ( this->empty() ) return true;
	if ( this->getTerms() > 1 ) return false;
	for ( i = 0; i < this->inputs(); i++ )
		if ( !data.front().isLit_dc(i) ) return false;
	return true;
}

bool SoPLA::isBuffer(void) const
// is the PLA a buffer?
{
	if ( this->empty() ) return false;
	if ( this->getTerms() != 1 ) return false;
	if ( this->inputs() != 1 ) return false;
	if ( data.front().isLit_1(0) && type ) return true;
	if ( data.front().isLit_0(0) && !type ) return true;
	return false;
}

bool SoPLA::isInverter(void) const
// is the PLA a inverter?
{
	if ( this->empty() ) return false;
	if ( this->getTerms() != 1 ) return false;
	if ( this->inputs() != 1 ) return false;
	if ( data.front().isLit_0(0) && type ) return true;
	if ( data.front().isLit_1(0) && !type ) return true;
	return false;
}

bool SoPLA::isTrivial(void) const
// is the PLA constant, buffer or inverter?
{
	return ( this->isConstant() || this->isBuffer() || this->isInverter() );
}

bool SoPLA::isAND_OR(void)
// is the PLA AND-OR (i.e. not only one AND or one OR gate)
{
	return ( this->inputs() > 1 && this->getTerms() > 1 && this->getLiterals() > this->inputs() );
}

Literal SoPLA::Simulate( const Cube &t ) const
// simulates a vector
{
	list<Cube>::const_iterator ti;
	Literal res;

	for ( ti = data.begin(); ti != data.end(); ++ti ) {
		res = ti->Simulate(t);
		if ( res != ZERO ) {
			if ( res == DC ) return DC;
			if ( res == ONE && type ) return ONE;
			else return ZERO;
		}
	}
	if ( type ) return ZERO;
	else return ONE;
}

bool SoPLA::RecognizeFunction( const string &n, const unsigned int _inps )
// tries to recognize the PLA function according to the gate name
// resurns true, if recognized
{
	Cube t;
	unsigned int i;
	unsigned long counter, endval;
	bool par;
	string s;
	
	if ( _inps != this->inps && this->inps != 0 ) throw RuntimeException( "SoPLA::RecognizeFunction: incorrect number of inputs" );
	this->gatename = n;
	to_upper(this->gatename);
	clearData();
	inps = _inps;
	if ( this->gatename == "AND" ) {
		t.assign( inps, ONE );
		this->AppendTerm(t);
		this->type = true;
	}
	else if ( this->gatename == "NAND" ) {
		t.assign( inps, ONE );
		this->AppendTerm(t);
		this->type = false;
	}
	else if ( this->gatename == "OR" ) {
		t.assign( inps, ZERO );
		this->AppendTerm(t);
		this->type = false;
	}
	else if ( this->gatename == "NOR" ) {
		t.assign( inps, ZERO );
		this->AppendTerm(t);
		this->type = true;
	}
	else if ( this->gatename == "OR" ) {
		t.assign( inps, ZERO );
		this->AppendTerm(t);
		this->type = false;
	}
	else if ( this->gatename == "XOR" ) {
		endval = (unsigned long)(1 << inps);
		for ( counter = 1; counter < endval; counter++ ) {
			s = int2bin( counter, inps );
			par = true;
			for ( i = 0; i < inps; i++ )
				if  ( s[i] == '1' ) par = !par;
			if ( !par ) this->AppendTerm(s);
		}
		this->type = true;
	}
	else if ( this->gatename == "XNOR" ) {
		endval = (unsigned long)(1 << inps);
		for ( counter = 0; counter < endval; counter++ ) {
			s = int2bin( counter, inps );
			par = true;
			for ( i = 0; i < inps; i++ )
				if  ( s[i] == '1' ) par = !par;
			if ( par ) this->AppendTerm(s);
		}
		this->type = true;
	}
	else if ( this->gatename == "BUF" || this->gatename == "BUFF" ) {
		t.assign(1, ONE);
		this->AppendTerm(t);
		this->type = true;
	}
	else if ( this->gatename == "INV" || this->gatename == "NOT" ) {
		t.assign(1, ZERO);
		this->AppendTerm(t);
		this->type = true;
	}
	else if ( this->gatename == "ZERO" ) {
		t.assign(1, DC);
		this->AppendTerm(t);
		this->type = false;
	}
	else if ( this->gatename == "ONE" ) {
		t.assign(1, DC);
		this->AppendTerm(t);
		this->type = true;
	}
	else return false;
	return true;
}